<?php
/**
* GeSHiExtreme
*
* version: 1.0-r209
*
* Copyright (c) 2008 Mark A. Taff <marktaff@comcast.net>
*
* Licensed under the GNU GPL, version 2, or any later version
* approved by Open Source Matters.
*
* Revised version of the Joomla GeshiPlus 1.0 plugin
*   ++ allows use of an alternate tag in Joomla content to allow geshi
*      markup in content that will be loaded and saved by editors.
*   ++ able to include an entire external file.
*   ++ permits reading in code between specific lines in an external file
*   ++ permits reading in all the code between two arbitray markers
*      in an external file
*   ++ permits using GeSHi's fancy line numbers
*   ++ permit starting line numbering from any number (N.B. Not
*      XHTML Strict compliant, thanks to IE's lack up support for CSS2)
*
* Prior copyright notices:
*
*   Copyright 2008 Hugo Jackson. All rights reserved.
*   This is a derivative work. Portion are copyright:
*       Open Source Matters
*   and
*       Fiji Web Design (www.fijiwebdesign.com)
*   This work is licensed under the GNU/GPL license.
*   This version may have been modified pursuant
*   to the GNU General Public License, and as distributed it includes or
*   is derivative of works licensed under the GNU General Public License or
*   other free or open source software licenses.
*/

// no direct access
defined( '_JEXEC' ) or die( 'Restricted access' );

jimport( 'joomla.event.plugin' );

/** Use <pre> tags */
define('OPTION_USE_PRE_TAG', 0);
/** Use {geshi}{/geshi} format tags */
define('OPTION_USE_GESHI_TAG', 1);

/** Do not number lines */
define('OPTION_NO_LINE_NUMBERS', 0);
/** Use normal line numbers */
define('OPTION_NORMAL_LINE_NUMBERS', 1);
/** Use fancy line numbers */
define('OPTION_FANCY_LINE_NUMBERS', 2);

/** Exclude tags themselves in tag mode */
define('OPTION_EXCLUDE_TAGS', 0);
/** Include tags themselves in tag mode */
define('OPTION_INCLUDE_TAGS', 1);


/**
* GeSHi Extreme Syntax Highlighting Plugin
*/
class plgContentGeshiExtreme extends JPlugin {

    /*
    *   Properties...
    */
    var $tagStyle    = null;        // Style of geshi tag to use
    var $lang        = null;        // Highlighting language
    var $lines       = null;        // Line numbering style
    var $fancyNth    = null;        // Make every nth line fancy
    var $src         = null;        // Absolute path to external source file
    var $beginLine   = null;        // Line to begin grabbing code from (inclusive)
    var $endLine     = null;        // Line the end grabbing code from (inclusive)
    var $initial     = null;        // Positive integer to start counting lines from
    var $beginTag    = null;        // Arbitrary text to begin grabbing text from
    var $endTag      = null;        // Arbitrary text to end grabbing text at
    var $includeTag  = null;        // Include/exclude arbitrary tags in output
    var $normalStyle = null;        // CSS style properties applied to normal lines
    var $fancyStyle  = null;        // CSS style properties applied to fancy lines
    var $text        = null;        // The text to highlight

    /**
     * ctor
     */
    function plgContentGeshiExtreme( &$subject )
    {
        parent::__construct( $subject );
        $this->_plugin =& JPluginHelper::getPlugin( 'content', 'geshiextreme' );
        $this->_params = new JParameter( $this->_plugin->params );

    } // End function plgContentGeshiExtreme()




   /**
    * Slot called when onPrepareContent event fires
    * @param array Unknown
    * @param array Unknown
    * @param integer Unknown
    * @return boolean A bool indicating success
    */
    function onPrepareContent( &$row, &$params, $page=0 )
    {
        $GLOBALS['_MAMBOT_GESHI_PARAMS'] =& $params;

        // Bail early if no row text
        if ($row->text == "") { return true;}

        // Initialize properties
        $this->initProperties();

        $regex = $this->buildRegexPattern();

        // Replace the text
        $row->text = preg_replace_callback( $regex, array($this, 'replacer'), $row->text );

        return true;
    } // End function onPrepareContent()




   /**
    * Reset properties to null
    */
    function initProperties()
    {
        // Reset properties
        $this->tagStyle     = null;
        $this->lang         = null;
        $this->lines        = null;
        $this->fancyNth     = null;
        $this->src          = null;
        $this->beginLine    = null;
        $this->endLine      = null;
        $this->initial      = null;
        $this->beginTag     = null;
        $this->endTag       = null;
        $this->includeTag   = null;
        $this->normalStyle  = null;
        $this->fancyStyle   = null;
        $this->text         = null;

        // Read default options
        $this->tagStyle      = $this->_params->get( 'option_tag_style' );
        $this->lang          = $this->_params->get( 'option_language' );
        $this->lines         = $this->_params->get( 'option_numbering_style' );
        $this->fancyNth      = $this->_params->get( 'option_fancy_nth_lines' );
        $this->includeTag    = $this->_params->get( 'option_include_tags' );

        /** @bug    Joomla! can't put # in a plugin parameter text field or description
         *          so fixCssHexadecimal() adds pound signs as needed to style info.
         */
        $this->normalStyle   = $this->_params->get( 'option_normal_style' );
        $this->fancyStyle    = $this->_params->get( 'option_fancy_style' );

    } // End function initProperties()




   /**
    * Replace the text
    * @param array An array of matches (see preg_match_all)
    * @return string String containing the replacement text
    */
    function replacer( &$matches)
    {
        /* Any GeSHi command that begins with a doubled brace or
         * angle bracket should not be run through GeSHi.  This
         * facilitates writing tutorials or examples on using
         * GeSHiExtreme
         */
        if ($matches[1] == "{{")
        {
            $this->text = str_replace('{{', '{', "<pre>".$matches[0]."</pre>");
            return $this->text;
        }
        elseif ($matches[1] == "<<")
        {
            $this->text = "<pre>".htmlspecialchars(str_replace('<<', '<', $matches[0]))."</pre>";
            return $this->text;
        }

        // Reset properties
        $this->initProperties();

        // Process global options and local attributes
        $this->processOptions($row, $matches);

        // Get the text to be highlighted
        $this->getCode($row);

        // Do the actual highlighting
        $this->highlight($row);
        return $this->text;
    }




   /**
    * Build the regex pattern that parses the GeSHi tags
    * @return string The regex pattern
    */
    function buildRegexPattern()
    {
        /** N.B. Complexity req'd to ensure arbitrary begintag or
        *  endtag attributes may contain braces or angle brackets.
        */

        $attribute = '\w+?\=\"[^"]+?\"';                    # A single attribute
        $br = '(?:\s*\\\\*(?:<br>|<br\s*/>|)\s*){0,}';      # Permit line breaks
        $attributes = "(?:$attribute$br){0,}";              # All atrributes
        $content = '(.*?)';                                 # Inline content
        $close = '{/geshi}';

        if( $this->tagStyle == OPTION_USE_GESHI_TAG )
        {
            $open = '({{|{)\s*geshi\s*xml:';                # Match start of opening geshi tag & `xml` namespace
            $close = '{/geshi}';
            $regex = "%$open($attributes)([}])$content$close%sx";
        }
        else
        {
            $open = '(<<|<)\s*pre\s*xml:';                  # Match start of opening geshi tag & `xml` namespace
            $close = '</pre>';
            $regex = "%$open($attributes)([>])$content$close%sx";
        }
        return $regex;
    } // End function buildRegexPattern()




   /**
    * Ensures every hexadecimal number is prepended with a pound sign
    * @param string The string to fix hexadecimal numbers of
    * @param string The corrected css style
    */
    function fixCssHexadecimal( $style )
    {
        // This pattern isn't perfect, but it might be close enough
        $color = '( [[:xdigit:]]{3} | [[:xdigit:]]{6} )';
        $pattern = "%(\s|:)$color(\s|;)%sx";
        $style = preg_replace($pattern, "$1#$2$3", $style);
        return $style;

    } // End function fixCssHexadecimal()




   /**
    * Processes global options and GeSHi tag attributes
    * @param array Unknown
    * @param array An array of matches (see preg_match_all)
    */
    function processOptions( &$row, &$matches )
    {
        $params =& $GLOBALS['_MAMBOT_GESHI_PARAMS'];

        jimport('geshi.geshi');
        jimport('domit.xml_saxy_shared');

        // Debug
        if (false) {for ($i=0; $i<7; ++$i) {echo "Matches[$i]: ". $matches[$i] ."<br />\n";}}

        // Strip any html breaks from the attribute string
        $attributes = str_replace("<br />", " ", $matches[2]);
        $attributes = str_replace("<br>", " ", $attributes);

        $args = SAXY_Parser_Base::parseAttributes( $attributes );
        $this->text = $matches[4];

        // Get attributes from geshi tag
        $this->lang        = JArrayHelper::getValue( $args, 'lang', $this->lang );
        $this->lines       = JArrayHelper::getValue( $args, 'lines', $this->lines );
        $this->fancyNth    = JArrayHelper::getValue( $args, 'interval', $this->fancyNth );
        $this->includeTag  = JArrayHelper::getValue( $args, 'includetag', $this->includeTag );

        $this->src         = JArrayHelper::getValue( $args, 'src', '' );
        $this->beginLine   = JArrayHelper::getValue( $args, 'beginline', '' );
        $this->endLine     = JArrayHelper::getValue( $args, 'endline', '' );
        $this->initial     = JArrayHelper::getValue( $args, 'initial', 1 );
        $this->beginTag    = JArrayHelper::getValue( $args, 'begintag', '' );
        $this->endTag      = JArrayHelper::getValue( $args, 'endtag', '' );
        $this->normalStyle = JArrayHelper::getValue( $args, 'normalstyle', $this->normalStyle );
        $this->fancyStyle  = JArrayHelper::getValue( $args, 'fancystyle', $this->fancyStyle );

        // Add missing pound signs (#)
        $this->normalStyle = $this->fixCssHexadecimal($this->normalStyle);
        $this->fancyStyle  = $this->fixCssHexadecimal($this->fancyStyle);

        // Validate attribute values
        if     ($this->lines == "false") {$this->lines = OPTION_NO_LINE_NUMBERS;}
        elseif ($this->lines == "true")  {$this->lines = OPTION_NORMAL_LINE_NUMBERS;}
        elseif ($this->lines == "fancy") {$this->lines = OPTION_FANCY_LINE_NUMBERS;}
        else   {$this->lines = $this->_params->get( 'option_numbering_style' );}

        if     ($this->includeTag == "false") {$this->includeTag = OPTION_EXCLUDE_TAGS;}
        elseif ($this->includeTag == "true")  {$this->includeTag = OPTION_INCLUDE_TAGS;}
        else   {$this->includeTag = $this->_params->get( 'option_include_tags' );}

        if ($this->initial < 0) {$this->initial = 0;}
        if ($this->fancyNth < 0) {$this->fancyNth = 5;}

    } // End function processOptions()




   /**
    * Gets the code to be highlighted by GeSHi
    * @param array Unknown
    */
    function getCode( &$row )
    {
        if (file_exists($this->src))
        {
            // Set operating mode
            if ($this->endLine > $this->beginLine) {$mode = "lines";}
            elseif (($this->beginTag != "") && ($this->endTag != "")) {$mode = "tags";}
            else {$mode = "slurp";}

            switch ($mode)
            {
                // Slurp in entire file
                case "slurp":
                    $this->text = file_get_contents($this->src);
                    break;

                // Grab a range of lines
                case "lines":
                    $currentLine = 1;
                    $this->text = "";
                    $handle = @fopen($this->src, "r");
                    if ($handle)
                    {
                        while (!feof($handle))
                        {
                            if (($currentLine >= $this->beginLine) && ($currentLine <= $this->endLine))
                            {
                                $this->text .= fgets($handle, 4096);
                            }
                            else {fgets($handle, 4096);}
                            $currentLine++;
                        }
                        fclose($handle);
                    }
                    else {$this->text = "Error: no file handle.";}
                    break;

                // Grab everything between arbitrary tags
                case "tags":
                        $haystack = file_get_contents($this->src);
                        $this->beginTag = preg_quote($this->beginTag, "/");
                        $this->endTag = preg_quote($this->endTag, "/");
                        if ($this->includeTag)
                        {
                            $pattern = '/('.$this->beginTag.'.*?'.$this->endTag.')/s';
                        }
                        else
                        {
                            $pattern = '/'.$this->beginTag.'(.*?)'.$this->endTag.'/s';
                        }
                        preg_match($pattern, $haystack, $found);
                        $this->text = $found[1];
                    break;

                default:
                    $this->text = "Ooops.  I couldn't determine what mode to operate in.";
            } // End switch ($mode)

        } // End if (file_exists($src))

        $html_entities_match = array( "|\<br \/\>|", "#<#", "#>#", "|&#39;|", '#&quot;#', '#&nbsp;#' );
        $html_entities_replace = array( "\n", '<', '>', "'", '"', ' ' );

        $this->text = preg_replace( $html_entities_match, $html_entities_replace, $this->text );

        $this->text = str_replace('<', '<', $this->text);
        $this->text = str_replace('>', '>', $this->text);

        $this->text = str_replace( "\t", '  ', $this->text );

    } // End function getCode()




   /**
    * Highlights the text with GeSHi
    * @param array An array of matches (see preg_match_all)
    */
    function highlight( &$row )
    {
        $geshi = new GeSHi( $this->text, $this->lang );

        // Determine numbering type & apply CSS styles
        if ($this->lines == OPTION_NORMAL_LINE_NUMBERS)
        {
            $lines = GESHI_NORMAL_LINE_NUMBERS;
            $geshi->set_line_style($this->normalStyle, true );
        }
        elseif ($this->lines == OPTION_FANCY_LINE_NUMBERS)
        {
            $lines = GESHI_FANCY_LINE_NUMBERS;
            $geshi->set_line_style( $this->normalStyle, $this->fancyStyle, true );
        }
        else
        {
            $lines = GESHI_NO_LINE_NUMBERS;
        }

        $geshi->enable_line_numbers( $lines, $this->fancyNth );

        // Start counting at specific number. N.B. This is *NOT*
        // valid XHMTL Strict, so do not use this method if
        // that is important to you.
        if ($this->initial > 0) {$geshi->start_line_numbers_at($this->initial);}

        // Now build the syntax highlighted version of the text
        $this->text = $geshi->parse_code();

    } // End function highlight()


} // End class plgContentGeshiExtreme